/*
 * Copyright 1996-2004 by Hans Reiser, licensing governed by
 * reiserfsprogs/README
 */

/* nothing abount reiserfs here */

#ifndef TIME_H
#define TIME_H

#include <time.h>

time_t parse_time(char *str);

#endif /* TIME_H */
